/*
 * bqfh.s.c
 * Copyright (C) 2009-2011, 2018, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bqfx-defs.h>
#include <bqfx-inter.h>
#include <bqfx-names.h>
#include <bqfx-types.h>

#define posh(___s, ___n) \
    (___s) = (char *) (___s) - (___n)

#define rule(___s, ___n, ___t) \
    ((char *) (___s) + (___n) * (___t))

#define skip(___s, ___n) \
    (___s) = (char *) (___s) + (___n)

#undef Q
#define Q(e)				((integral_q) (e))

#define bqfset(bqfset)			((struct bqfset_type *) (bqfset))

static int line(void *, unsigned, unsigned, void *, int (*)(void *, void *));
static int slip(void *, void *, unsigned, unsigned, unsigned, void *, void *,
		int (*)(void *, void *));

static int
line(void *node, unsigned size, unsigned miss, void *back,
     int (*call)(void *, void *))
{
    int delete;
    unsigned note;

    note = *(integral_q *) node - miss;
    if (note) {
	node = rule(node, size, miss);
	do {
	    skip(node, size);
	    note--;
	    delete = call(back, node);
	    if (delete) {
		break;
	    }
	} while (note);
    } else {
	delete = 0;
    }

    return delete;
}


static int
slip(void *bqfset, void *node, unsigned rate, unsigned size, unsigned miss,
     void *aime, void *back, int (*call)(void *, void *))
{
    int delete;
    unsigned note;
    void **text;

    text = (void **) ((char *) node + size * integral_q_bits_and_half);

    text += miss;

    note = *(integral_q *) node - miss;

    node = rule(node, size, miss);

    rate--;

    if (rate) {
	delete = _libx1f4l2_fxpi_bqfset
	    (bqfset, *text, rate, size, aime, back, call);
	if (delete) {
	} else {
	    while (note) {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		    break;
		} else {
		    text++;
		    delete = _libx1f4l2_fpxx_bqfset
			(*text, rate, size, back, call);
		    if (delete) {
			break;
		    }
		}

		note--;
	    }
	}
    } else {
	delete = _libx1f4l2_fxpe_bqfset(bqfset, *text, size, aime, back, call);
	if (delete) {
	} else {
	    while (note) {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		    break;
		} else {
		    text++;
		    delete = _libx1f4l2_flxx_bqfset(*text, size, back, call);
		    if (delete) {
			break;
		    }
		}

		note--;
	    }
	}
    }

    return delete;
}


int
x1f4_fxpd_bqfset(void *bqfset, void *aime, void *back,
		 int (*call)(void *, void *))
{
    int delete;
    unsigned copy, rate, size;
    void *node;

    node = bqfset(bqfset)->link_a.fpnews.node;

    size = bqfset(bqfset)->link_a.fpnews.link_v.size;

    if (bqfset(bqfset)->link_f.look(aime, node, &copy)) {
	void *last;

	last = rule(node, size, copy);

	copy--;
	while (copy) {
	    posh(last, size);
	    if (bqfset(bqfset)->link_f.fare(aime, last)) {
		break;
	    } else {
		copy--;
	    }
	}
    }

    rate = bqfset(bqfset)->link_a.fpnews.link_v.rate;
    if (rate) {
	delete = slip(bqfset, node, rate, size, copy, aime, back, call);
    } else {
	delete = line(node, size, copy, back, call);
    }

    return delete;
}
