/*
 * bqfi.p.c
 * Copyright (C) 2009-2011, 2018, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bqfx-defs.h>
#include <bqfx-inter.h>
#include <bqfx-names.h>
#include <bqfx-types.h>

#define posh(___s, ___n) \
    (___s) = (char *) (___s) - (___n)

#define rule(___s, ___n, ___t) \
    ((char *) (___s) + (___n) * (___t))

#define skip(___s, ___n) \
    (___s) = (char *) (___s) + (___n)

#undef Q
#define Q(e)				((integral_q) (e))

#define bqfset(bqfset)			((struct bqfset_type *) (bqfset))

static int line(void *, unsigned, unsigned, unsigned, void *,
		int (*)(void *, void *));
static int slip(void *, void *, unsigned, unsigned, unsigned, void *, unsigned,
		void *, void *, int (*)(void *, void *));

static const struct ordain_type uodale = {
/* *INDENT-OFF* */
    _libx1f4l2_xxoe_bqfset,
    _libx1f4l2_fxpe_bqfset,
    _libx1f4l2_fxpi_bqfset,
    _libx1f4l2_xxul_bqfset,
    _libx1f4l2_lxpe_bqfset,
    _libx1f4l2_lxpi_bqfset,
    _libx1f4l2_uope_bqfset,
/* *INDENT-ON* */
};

static int
line(void *node, unsigned size, unsigned cast, unsigned copy, void *back,
     int (*call)(void *, void *))
{
    int delete;

    copy -= cast;
    if (copy) {
	node = rule(node, size, cast);
	do {
	    skip(node, size);
	    copy--;
	    delete = call(back, node);
	    if (delete) {
		break;
	    }
	} while (copy);
    } else {
	delete = 0;
    }

    return delete;
}


static int
slip(void *bqfset, void *node, unsigned rate, unsigned size, unsigned cast,
     void *lead, unsigned copy, void *ever, void *back,
     int (*call)(void *, void *))
{
    int delete;
    void **text;

    text = (void **) ((char *) node + size * integral_q_bits_and_half);

    text += cast;

    node = rule(node, size, cast);

    if (rate) {
	delete = _libx1f4l2_fxpi_bqfset
	    (bqfset, *text, rate, size, lead, back, call);
	if (delete) {
	} else {
	    copy -= cast;
	    copy--;
	    while (copy) {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		    break;
		} else {
		    text++;
		    delete = _libx1f4l2_fpxx_bqfset
			(*text, rate, size, back, call);
		    if (delete) {
			break;
		    }
		}

		copy--;
	    }

	    if (delete) {
	    } else {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		} else {
		    text++;
		    delete = _libx1f4l2_lxpi_bqfset
			(bqfset, *text, rate, size, ever, back, call);
		}
	    }
	}
    } else {
	delete = _libx1f4l2_fxpe_bqfset(bqfset, *text, size, lead, back, call);
	if (delete) {
	} else {
	    copy -= cast;
	    copy--;
	    while (copy) {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		    break;
		} else {
		    text++;
		    delete = _libx1f4l2_flxx_bqfset(*text, size, back, call);
		    if (delete) {
			break;
		    }
		}

		copy--;
	    }

	    if (delete) {
	    } else {
		skip(node, size);
		delete = call(back, node);
		if (delete) {
		} else {
		    text++;
		    delete = _libx1f4l2_lxpe_bqfset
			(bqfset, *text, size, ever, back, call);
		}
	    }
	}
    }

    return delete;
}


int
x1f4_uopd_bqfset(void *bqfset, void *lead, void *ever, void *back,
		 int (*call)(void *, void *))
{
    int delete;
    unsigned cast, copy, rate, size;
    void *node;

    node = bqfset(bqfset)->link_a.fpnews.node;

    size = bqfset(bqfset)->link_a.fpnews.link_v.size;

    if (bqfset(bqfset)->link_f.look(lead, node, &cast)) {
	void *last;

	last = rule(node, size, cast);

	cast--;
	while (cast) {
	    posh(last, size);
	    if (bqfset(bqfset)->link_f.fare(lead, last)) {
		break;
	    } else {
		cast--;
	    }
	}
    }

    if (bqfset(bqfset)->link_f.look(ever, node, &copy)) {
	unsigned mind;
	void *next;

	next = rule(node, size, copy);

	mind = *(integral_q *) node;

	while (copy ^ mind) {
	    skip(next, size);
	    if (bqfset(bqfset)->link_f.fare(ever, next)) {
		break;
	    } else {
		copy++;
	    }
	}
    }

    rate = bqfset(bqfset)->link_a.fpnews.link_v.rate;
    if (rate) {
	rate--;
	if (copy ^ cast) {
	    delete = slip
		(bqfset, node, rate, size, cast, lead, copy, ever, back, call);
	} else {
	    void **text;

	    text = (void **) ((char *) node + size * integral_q_bits_and_half);
	    text += cast;
	    if (rate) {
		delete = _libx1f4l2_xxpi_bqfset
		    (bqfset, *text, rate, size, lead, ever, &uodale, back,
		     call);
	    } else {
		delete = _libx1f4l2_uope_bqfset
		    (bqfset, *text, size, lead, ever, back, call);
	    }
	}
    } else {
	delete = line(node, size, cast, copy, back, call);
    }

    return delete;
}
