/*
 * lxdash-i.q.c
 * Copyright (C) 2008-2016, 2018, 2019, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <bcount.h>
#include <d_fset.h>
#include <integral_q.h>
#include <lxdash-drive.h>
#include <lxdash-types.h>

#define __SINGLE__			1

#define effect(___s) \
    ((struct effect_type *) (___s))
#define seek(___s) \
    ((struct seek_type *) (___s))

typedef struct seek_type {
    X1f4_E4_C_MODE mode;
    unsigned ever;
#if __SINGLE__
#else
    void *data;
#endif				/* __SINGLE__ */
} seek_type;

static int slip_cast(void *, void *);
static int slip_data(void *, void *);
static int slip_line(void *, void *);
#if __SINGLE__
#else
static int slip_roll(void *, void *);
#endif				/* __SINGLE__ */

static int
slip_cast(void *seek, void *data)
{
    X1f4_E4_C_MODE mode;
    int delete;

    mode = effect(data)->mode;
    if (mode < seek(seek)->mode) {
	seek(seek)->ever++;
	delete = 0;
    } else {
	delete = 1;
	if (mode ^ seek(seek)->mode) {
	} else {
#if __SINGLE__
#else
	    seek(seek)->data = NULL;
#endif				/* __SINGLE__ */
	}
    }

    return delete;
}


static int
slip_data(void *seek, void *data)
{
    X1f4_E4_C_MODE mode;
    int delete;

    mode = effect(data)->mode;
    if (mode < seek(seek)->mode) {
	seek(seek)->ever++;
	delete = 0;
    } else {
	delete = 1;
	if (mode ^ seek(seek)->mode) {
	} else {
#if __SINGLE__
#else
	    seek(seek)->data = NULL;
#endif				/* __SINGLE__ */
	    seek(seek)->ever++;
	}
    }

    return delete;
}


static int
slip_line(void *seek, void *data)
{
    seek(seek)->ever++;

    return 0;
}


#if __SINGLE__
#else
static int
slip_roll(void *seek, void *data)
{
    unsigned ever;

    l2_q_bitcount(ever, *(integral_q *) data >> 1);
    seek(seek)->data = data;
    seek(seek)->ever += ever;

    return 0;
}
#endif				/* __SINGLE__ */


int
_libx1f4i0_lxdash_size_line(void *dxdash, void *lead, void *rail,
			    unsigned *count, int infer)
{
    X1f4_E4_C_MODE head, tail;
    struct dxdash_type *dxdash_data;
    struct seek_type seek;

    dxdash_data = dxdash;

    head = I_MODE(lead);

    tail = I_MODE(rail);

#if __SINGLE__
#else
    seek.data = NULL;
#endif				/* __SINGLE__ */
    seek.ever = 0;
    seek.mode = tail;

#if __SINGLE__
    if (tail < head) {
	_libx1f4i0_fxpd_lxfset(dxdash_data->dash, &head, &seek, slip_line);
	(infer ? _libx1f4i0_lxpd_lxfset : _libx1f4i0_llpd_lxfset)
	    (dxdash_data->dash, &tail, &seek, slip_line);
    } else {
	_libx1f4i0_fxpd_lxfset
	    (dxdash_data->dash, &head, &seek, infer ? slip_data : slip_cast);
    }
#else
    _libx1f4i0_fxrd_lxfset
	(dxdash_data->dash, &head, &seek, slip_data, slip_roll);
    if (seek.data) {
	integral_q news;
	struct effect_type *effect_data;
	unsigned ever;

	effect_data = seek.data;

	news = *(integral_q *) effect_data;
	news >>= 1;

	for (; news; news >>= 1) {
	    effect_data++;
	    if (news & 1) {
		if (tail < effect_data->mode) {
		    break;
		}
	    }
	}

	l2_q_bitcount(ever, news);
	seek.ever -= ever;
    }
#endif				/* __SINGLE__ */

    *count = seek.ever;

    return 0;
}
