/*
 * lxline-i.q.c
 * Copyright (C) 2008-2016, 2018, 2019, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <e4.h>
#include <integral_q.h>
#include <l_note.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

#define seek(___s) \
    ((struct seek_type *) (___s))

typedef struct seek_type {
    unsigned ever;
} seek_type;

static int tick_link(void *, void *);

static int
tick_link(void *seek, void *data)
{
    seek(seek)->ever++;

    return 0;
}


int
_libx1f4i0_lxline_size_line(void *lxline, void *dxline, void *lead, void *rail,
			    unsigned *count, int infer)
{
    char *head, *tail;
    int rule, status;
    void *half, *part;

    head = I_TEXT(lead);

    tail = I_TEXT(rail);

    rule = _x1f4_e4_text_diff(tail, head);

    if (rule < 0) {
	char *near;

	near = head;
	head = tail;
	tail = near; 
    }

    status = _libx1f4i0_lxline_mill_line
	(lxline, dxline(dxline), head, tail, &half, &part, -1 < rule);
    if (status) {
    } else {
	int excess;
	struct seek_type seek;
	void *line;

	line = dxline(dxline)->line;

	seek.ever = 0;
	if (rule < 0) {
	    _libx1f4i0_lime_sfnote(line, &seek, tick_link);
	    _libx1f4i0_lime_sfnote(part, &seek, tick_link);
	} else {
	    _libx1f4i0_lime_sfnote(half, &seek, tick_link);
	}

	status = _libx1f4i0_join_sfnote(line, half);
	if (status) {
	} else {
	    status = _libx1f4i0_join_sfnote(line, part);
	    if (status) {
	    } else {
		*count = seek.ever;
	    }
	}

	excess = _libx1f4i0_fini_sfnote(&half);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}

	excess = _libx1f4i0_fini_sfnote(&part);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}
