/*
 * lxline-i.h.c
 * Copyright (C) 2008-2016, 2018, 2019, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <e4.h>
#include <l_note.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

int
_libx1f4i0_lxline_ever_data(void *line, void *data)
{
    int link, status;

    do {
	struct direct_type *direct_data;
	struct either_type *either_data;

	either_data = line;

	direct_data = either_data->either;

	direct_data->link = effect_link(data);
	direct_data->type = effect_type(data);

	direct_data->fxdata = *(struct fxdata_type *) effect_data(data);

	link = direct_data->link;
	if (~link) {
	    struct lxtext_type *lxtext_data;

	    lxtext_data = either_data->lxtext_data + link;

	    if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
		status = lxtext_data->lxtype.node
		    (lxtext_data->lxtype.context,
		     (void *) &direct_data->fxdata,
		     I_USER(&direct_data->fxdata), either_data->excase_data);
		if (status) {
		    break;
		} else {
		    either_data->roll = 1;
		}
	    } else {
		int (*copy)(void *, void *, void **, const void *,
			    const struct x1f4_nodelink_type *);

		copy = lxtext_data->lxtype.copy;
		if (copy) {
		    if (lxtext_data->lxtype.code == X1f4_E4_TEXT) {
			char *down;

			direct_data->link = -1;

			down = I_TEXT(&direct_data->fxdata);
			status = _libx1f4i0_lxline_rule_text
			    (&direct_data->fxdata, down, strlen(down),
			     either_data);
			if (status) {
			    break;
			} else {
			}
		    } else {
			status = copy
			    (lxtext_data->lxtype.context, NULL,
			     (void *) &direct_data->fxdata,
			     I_USER(&direct_data->fxdata),
			     either_data->excase_data);
			if (status) {
			    break;
			} else {
			    either_data->roll = 1;
			}
		    }
		} else {
		    status = 0;
		}
	    }
	} else {
	    status = 0;
	}

	either_data->either = direct_data + 1;
    } while (0);

    return status;
}
