/*
 * lxline-i.w.c
 * Copyright (C) 2011-2017, 2022, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

int
_libx1f4i0_lxline_rule_data(void *line, void *data, const char *text,
			    unsigned size)
{
    int link, status;

    do {
	struct affect_type *affect_data;
	struct either_type *either_data;

	either_data = line;

	affect_data = either_data->either;

	status = _libx1f4i0_lxline_rule_text
	    (&affect_data->fxtext, text, size, either_data);
	if (status) {
	    break;
	}

	affect_data->link = effect_link(data);
	affect_data->type = effect_type(data);

	affect_data->fxdata = *(struct fxdata_type *) (effect_data(data));

	link = affect_data->link;
	if (~link) {
	    struct lxtext_type *lxtext_data;

	    lxtext_data = either_data->lxtext_data + link;

	    if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
		status = lxtext_data->lxtype.node
		    (lxtext_data->lxtype.context,
		     (void *) &affect_data->fxdata,
		     I_USER(&affect_data->fxdata), either_data->excase_data);
		if (status) {
		    break;
		} else {
		    either_data->roll = 1;
		}
	    } else {
		int (*copy)(void *, void *, void **, const void *,
			    const struct x1f4_nodelink_type *);

		copy = lxtext_data->lxtype.copy;
		if (copy) {
		    if (lxtext_data->lxtype.code == X1f4_E4_TEXT) {
			char *down;

			affect_data->link = -1;

			down = I_TEXT(&affect_data->fxdata);
			status = _libx1f4i0_lxline_rule_text
			    (&affect_data->fxdata, down, strlen(down),
			     either_data);
			if (status) {
			    break;
			} else {
			}
		    } else {
			status = copy
			    (lxtext_data->lxtype.context, NULL,
			     (void *) &affect_data->fxdata,
			     I_USER(&affect_data->fxdata),
			     either_data->excase_data);
			if (status) {
			    break;
			} else {
			    either_data->roll = 1;
			}
		    }
		} else {
		}
	    }
	} else {
	}

	either_data->either = affect_data + 1;
    } while (0);

    return status;
}
