/*
 * lxvery-i.q.c
 * Copyright (C) 2008-2016, 2018, 2019, 2024, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lxvery-config.h>

#include <stddef.h>

#include <bcount.h>
#include <d_fset.h>
#include <integral_q.h>
#include <lxvery-drive.h>
#include <lxvery-types.h>

#define __SINGLE__			1

#define dxvery(line) \
    ((struct dxvery_type *) (line))

#define seek(___s) \
    ((struct seek_type *) (___s))

typedef struct seek_type {
#if __SINGLE__
#else
    F_TYPE mode;
#endif				/* __SINGLE__ */
    unsigned ever;
#if __SINGLE__
#else
    void *data;
#endif				/* __SINGLE__ */
} seek_type;

#if __SINGLE__
#else
static int slip_data(void *, void *);
#endif				/* __SINGLE__ */
static int slip_line(void *, void *);
#if __SINGLE__
#else
static int slip_roll(void *, void *);
#endif				/* __SINGLE__ */

#if __SINGLE__
#else
static int
slip_data(void *seek, void *data)
{
    F_TYPE mode;
    int delete;

    mode = *(X1f4_E4_C_MODE *) data;
    if (mode < seek(seek)->mode) {
	seek(seek)->ever++;
	delete = 0;
    } else {
	delete = 1;
	if (mode ^ seek(seek)->mode) {
	} else {
	    seek(seek)->data = NULL;
	    seek(seek)->ever++;
	}
    }

    return delete;
}
#endif				/* __SINGLE__ */


static int
slip_line(void *seek, void *data)
{
    seek(seek)->ever++;

    return 0;
}


#if __SINGLE__
#else
static int
slip_roll(void *seek, void *data)
{
    unsigned ever;

    l2_q_bitcount(ever, *(integral_q *) data >> 1);
    seek(seek)->data = data;
    seek(seek)->ever += ever;

    return 0;
}
#endif				/* __SINGLE__ */


int
_libx1f4i0_lxvery_size_line(void *dxvery, void *lead, void *rail,
			    unsigned *count, int infer)
{
    F_TYPE head, tail;
    struct seek_type seek;

    head = I_MODE(lead);
#if __VERY_PSET__
    head += ~(X1f4_E4_C_BILL) 0 >> 1;
#endif				/* __VERY_PSET__ */

    tail = I_MODE(rail);
#if __VERY_PSET__
    tail += ~(X1f4_E4_C_BILL) 0 >> 1;
#endif				/* __VERY_PSET__ */

#if __SINGLE__
#else
    seek.data = NULL;
#endif				/* __SINGLE__ */
    seek.ever = 0;
#if __SINGLE__
#else
    seek.mode = tail;
#endif				/* __SINGLE__ */

#if __SINGLE__
    if (tail < head) {
	_libx1f4i0_fxpd_lxfset(dxvery(dxvery)->dash, &head, &seek, slip_line);
	(infer ? _libx1f4i0_lxpd_lxfset : _libx1f4i0_llpd_lxfset)
	    (dxvery(dxvery)->dash, &tail, &seek, slip_line);
    } else {
	(infer ? _libx1f4i0_uopd_lxfset : _libx1f4i0_ulpd_lxfset)
	    (dxvery(dxvery)->dash, &head, &tail, &seek, slip_line);
    }
#else
    _libx1f4i0_fxrd_lxfset
	(dxvery(dxvery)->dash, &head, &seek, slip_data, slip_roll);
    if (seek.data) {
	integral_q *line, news;
	unsigned ever;

	line = seek.data;

	news = *line;
	news >>= 1;

	for (; news; news >>= 1) {
	    line++;
	    if (news & 1) {
		if ((F_TYPE) tail < *(F_TYPE *) line) {
		    break;
		}
	    }
	}

	l2_q_bitcount(ever, news);
	seek.ever -= ever;
    }
#endif				/* __SINGLE__ */

    *count = seek.ever;

    return 0;
}
